from typing import Callable, Generator

from reling.config import MAX_SCORE
from reling.db.enums import ContentCategory
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.types import DialogueExchangeData, Promise
from reling.utils.feeders import CharFeeder
from reling.utils.values import coalesce
from .types import ExchangeWithTranslation, ExplanationRequest, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'build_dialogue_explainer',
    'build_text_explainer',
]

MISTAKE_THRESHOLD_RATIO = 0.5


def if_not(text: str | None, until_equals: str) -> str | None:
    """Return the text if it does not equal the specified value, otherwise return None."""
    return text if text != until_equals else None


def explain_structure(sentence: str, language: Language) -> list[str]:
    """Generate a prompt section for explaining the structure of a translation in the specified language."""
    return [
        f'The last line is translated into {language.name} as follows:',
        f'"""{sentence}"""',
        f'Explain the structure of this translation to a learner of {language.name}.',
    ]


def explain_mistakes(sentence: str, corrected: str, language: Language) -> list[str]:
    """Generate a prompt section for explaining why a corrected translation is preferred over the provided one."""
    return [
        f'A learner of {language.name} translated the last line as follows:',
        f'"""{sentence}"""',
        f'Explain why the following translation is more appropriate:'
        f'"""{corrected}"""',
    ]


def explain_difference(sentence: str, alternative: str, language: Language) -> list[str]:
    """Generate a prompt section for explaining differences between a provided and alternative translation."""
    return [
        f'A learner of {language.name} translated the last line as follows:',
        f'"""{sentence}"""',
        f'Explain whether there is any difference from the alternative translation below '
        f'and which option is more appropriate:',
        f'"""{alternative}"""',
    ]


def explain_source_structure(sentence: str) -> list[str]:
    """Generate a prompt section for explaining the structure of a sentence in the source language."""
    return [
        f'Explain the grammatical and structural elements of the last line ("""{sentence}""") to a language learner.',
    ]


def do_explain(
        gpt: Promise[GPTClient],
        category: ContentCategory,
        initial_sentences: list[str],
        provided: str,
        original: str,
        result: ScoreWithSuggestion,
        source_language: Language,
        target_language: Language,
        explain_source: bool,
) -> Generator[str, None, None]:
    """Generate an explanation for translations based on user-provided and corrected data."""
    return gpt().ask(
        prompt='\n'.join([
            f'Here is the beginning of a {category.value} in {source_language.name}:',
            f'',
            *initial_sentences,
            f'',
            *(
                explain_source_structure(initial_sentences[-1])
                if explain_source else
                explain_difference(
                    sentence=provided,
                    alternative=alternative,
                    language=target_language,
                )
                if result.score == MAX_SCORE and (alternative := coalesce(
                    if_not(result.suggestion, provided),
                    if_not(original, provided),
                )) else
                explain_mistakes(
                    sentence=provided,
                    corrected=result.suggestion or original,
                    language=target_language,
                )
                if MISTAKE_THRESHOLD_RATIO * MAX_SCORE < result.score < MAX_SCORE else
                explain_structure(
                    sentence=result.suggestion or original,
                    language=target_language,
                )
            ),
            f'',
            f'Be concise. Do not repeat the sentences; start directly with the answer.',
        ]),
        creative=False,
        feeder_type=CharFeeder,
        auto_normalize=False,
    )


def build_text_explainer(
        gpt: Promise[GPTClient],
        sentences: list[SentenceWithTranslation],
        original_translations: list[str],
        results: list[ScoreWithSuggestion],
        source_language: Language,
        target_language: Language,
) -> Callable[[ExplanationRequest], Generator[str, None, None]]:
    """Create a function to generate explanations for text translations."""
    def explain(request: ExplanationRequest) -> Generator[str, None, None]:
        index = request.sentence_index
        return do_explain(
            gpt=gpt,
            category=ContentCategory.TEXT,
            initial_sentences=[sentence.sentence for sentence in sentences[:index + 1]],
            provided=sentences[index].translation.text,
            original=original_translations[index],
            result=results[index],
            source_language=source_language,
            target_language=target_language,
            explain_source=request.source,
        )
    return explain


def build_dialogue_explainer(
        gpt: Promise[GPTClient],
        exchanges: list[ExchangeWithTranslation],
        original_translations: list[DialogueExchangeData],
        results: list[ScoreWithSuggestion],
        source_language: Language,
        target_language: Language,
) -> Callable[[ExplanationRequest], Generator[str, None, None]]:
    """Create a function to generate explanations for dialogue translations."""
    def explain(request: ExplanationRequest) -> Generator[str, None, None]:
        index = request.sentence_index
        return do_explain(
            gpt=gpt,
            category=ContentCategory.DIALOGUE,
            initial_sentences=[turn for exchange in exchanges[:index + 1] for turn in exchange.exchange.all()],
            provided=exchanges[index].user_translation.text,
            original=original_translations[index].user,
            result=results[index],
            source_language=source_language,
            target_language=target_language,
            explain_source=request.source,
        )
    return explain



from __future__ import annotations
from math import ceil, floor
from statistics import mean

from lcs2 import diff, lcs_indices, lcs_length

from reling.config import MAX_SCORE
from reling.utils.strings import tokenize

__all__ = [
    'calculate_diff_score',
    'format_average_score',
]

FUZZY_RATIO = 0.61


def calculate_mistake_score(mistakes: int) -> int:
    """Return the score based on the number of mistakes."""
    return floor(MAX_SCORE * ((1 - 1 / MAX_SCORE) ** mistakes))


def calculate_lcs_score(lcs_len: int, a_len: int, b_len: int) -> int:
    """Return the score based on the length of the longest common subsequence of two sequences."""
    if lcs_len == a_len == b_len:
        return MAX_SCORE
    else:
        return ceil(lcs_len / max(a_len, b_len) * (MAX_SCORE - 1))


def calculate_char_diff_score(a: str, b: str) -> int:
    """Return the score based on the longest common subsequence of two strings."""
    return calculate_lcs_score(lcs_length(a, b), len(a), len(b))


class FuzzyWord:
    """
    A class representing a word for fuzzy comparison.
    Two words are considered equal if they share at least FUZZY_RATIO of their characters, ignoring case.
    """
    _normalized: str

    def __init__(self, word: str) -> None:
        self._normalized = word.lower()

    def __eq__(self, other: FuzzyWord) -> bool:
        return (2 * lcs_length(self._normalized, other._normalized)
                >= (len(self._normalized) + len(other._normalized)) * FUZZY_RATIO)


def calculate_fuzzy_word_diff_score(a: str, b: str, cj: bool) -> int:
    """
    Calculate the score based on the total length of the longest common subsequences of characters
    within fuzzily aligned words from two strings.
    """
    a_words, b_words = (tokenize(sentence, punctuation=False, whitespace=False, cj=cj) for sentence in (a, b))
    a_len, b_len = (sum(map(len, words)) for words in (a_words, b_words))
    lcs_len = 0
    for a_index, b_index in lcs_indices(map(FuzzyWord, a_words), map(FuzzyWord, b_words)):
        lcs_len += lcs_length(a_words[a_index], b_words[b_index])
    return calculate_lcs_score(lcs_len, a_len, b_len)


def calculate_word_mistake_diff_score(a: str, b: str, cj: bool) -> int:
    """Return the score based on the number of mistakes computed from a word-level diff between two strings."""
    a_words, b_words = ([word.lower() for word in tokenize(sentence, punctuation=False, whitespace=False, cj=cj)]
                        for sentence in (a, b))
    a_diff: list[str] = []
    b_diff: list[str] = []
    mistakes = 0
    for a_tokens, b_tokens in diff(a_words, b_words):
        mistakes += max(len(a_tokens), len(b_tokens))
        a_diff.extend(a_tokens)
        b_diff.extend(b_tokens)
    mistakes -= lcs_length(a_diff, b_diff)
    return calculate_mistake_score(mistakes)


def calculate_diff_score(a: str, b: str) -> int:
    """Return the score based on the diff between two strings."""
    return min(
        calculate_char_diff_score(a, b),
        calculate_fuzzy_word_diff_score(a, b, cj=True),
        calculate_word_mistake_diff_score(a, b, cj=False),
    )


def format_average_score(scores: list[int]) -> str:
    return f'{mean(scores):.1f}'
